@@ -108,8 +108,8 @@ static double max_clock_error;
 
 #define NSEC_PER_SEC 1000000000
 
-static void
-calculate_sys_precision(void)
+static double
+measure_clock_precision(void)
 {
   struct timespec ts, old_ts;
   int iters, diff, best;
@@ -135,18 +135,7 @@ calculate_sys_precision(void)
 
   assert(best > 0);
 
-  precision_quantum = 1.0e-9 * best;
-
-  /* Get rounded log2 value of the measured precision */
-  precision_log = 0;
-  while (best < 707106781) {
-    precision_log--;
-    best *= 2;
-  }
-
-  assert(precision_log >= -30);
-
-  DEBUG_LOG(LOGF_Local, "Clock precision %.9f (%d)", precision_quantum, precision_log);
+  return 1.0e-9 * best;
 }
 
 /* ================================================== */
@@ -170,7 +159,16 @@ LCL_Initialise(void)
   current_freq_ppm = 0.0;
   temp_comp_ppm = 0.0;
 
-  calculate_sys_precision();
+  precision_quantum = CNF_GetClockPrecision();
+  if (precision_quantum <= 0.0)
+    precision_quantum = measure_clock_precision();
+
+  precision_quantum = CLAMP(1.0e-9, precision_quantum, 1.0);
+  precision_log = round(log(precision_quantum) / log(2.0));
+  /* NTP code doesn't support smaller log than -30 */
+  assert(precision_log >= -30);
+
+  DEBUG_LOG("Clock precision %.9f (%d)", precision_quantum, precision_log);
 
   /* This is the maximum allowed frequency offset in ppm, the time must
      never stop or run backwards */
@@ -185,13 +183,11 @@ LCL_Initialise(void)
 void
 LCL_Finalise(void)
 {
-  while (change_list.next != &change_list)
-    LCL_RemoveParameterChangeHandler(change_list.next->handler,
-                                     change_list.next->anything);
-
-  while (dispersion_notify_list.next != &dispersion_notify_list)
-    LCL_RemoveDispersionNotifyHandler(dispersion_notify_list.next->handler,
-                                      dispersion_notify_list.next->anything);
+  /* Make sure all handlers have been removed */
+  if (change_list.next != &change_list)
+    assert(0);
+  if (dispersion_notify_list.next != &dispersion_notify_list)
+    assert(0);
 }
 
 /* ================================================== */
@@ -358,12 +354,12 @@ LCL_ReadRawTime(struct timespec *ts)
 {
 #if HAVE_CLOCK_GETTIME
   if (clock_gettime(CLOCK_REALTIME, ts) < 0)
-    LOG_FATAL(LOGF_Local, "clock_gettime() failed : %s", strerror(errno));
+    LOG_FATAL("clock_gettime() failed : %s", strerror(errno));
 #else
   struct timeval tv;
 
   if (gettimeofday(&tv, NULL) < 0)
-    LOG_FATAL(LOGF_Local, "gettimeofday() failed : %s", strerror(errno));
+    LOG_FATAL("gettimeofday() failed : %s", strerror(errno));
 
   UTI_TimevalToTimespec(&tv, ts);
 #endif
@@ -426,7 +422,7 @@ clamp_freq(double freq)
   if (freq <= max_freq_ppm && freq >= -max_freq_ppm)
     return freq;
 
-  LOG(LOGS_WARN, LOGF_Local, "Frequency %.1f ppm exceeds allowed maximum", freq);
+  LOG(LOGS_WARN, "Frequency %.1f ppm exceeds allowed maximum", freq);
 
   return CLAMP(-max_freq_ppm, freq, max_freq_ppm);
 }
@@ -440,7 +436,7 @@ check_offset(struct timespec *now, double offset)
   if (UTI_IsTimeOffsetSane(now, -offset))
       return 1;
 
-  LOG(LOGS_WARN, LOGF_Local, "Adjustment of %.1f seconds is invalid", -offset);
+  LOG(LOGS_WARN, "Adjustment of %.1f seconds is invalid", -offset);
   return 0;
 }
 
@@ -546,7 +542,7 @@ LCL_ApplyStepOffset(double offset)
       return 0;
 
   if (!(*drv_apply_step_offset)(offset)) {
-    LOG(LOGS_ERR, LOGF_Local, "Could not step clock");
+    LOG(LOGS_ERR, "Could not step system clock");
     return 0;
   }
 
@@ -613,7 +609,7 @@ LCL_AccumulateFrequencyAndOffset(double dfreq, double doffset, double corr_rate)
 
   current_freq_ppm = clamp_freq(current_freq_ppm);
 
-  DEBUG_LOG(LOGF_Local, "old_freq=%.3fppm new_freq=%.3fppm offset=%.6fsec",
+  DEBUG_LOG("old_freq=%.3fppm new_freq=%.3fppm offset=%.6fsec",
       old_freq_ppm, current_freq_ppm, doffset);
 
   /* Call the system-specific driver for setting the frequency */
@@ -660,7 +656,7 @@ lcl_RegisterSystemDrivers(lcl_ReadFrequencyDriver read_freq,
 
   current_freq_ppm = (*drv_read_freq)();
 
-  DEBUG_LOG(LOGF_Local, "Local freq=%.3fppm", current_freq_ppm);
+  DEBUG_LOG("Local freq=%.3fppm", current_freq_ppm);
 }
 
 /* ================================================== */
@@ -684,7 +680,7 @@ LCL_MakeStep(void)
   if (!LCL_ApplyStepOffset(-correction))
     return 0;
 
-  LOG(LOGS_WARN, LOGF_Local, "System clock was stepped by %.6f seconds", correction);
+  LOG(LOGS_WARN, "System clock was stepped by %.6f seconds", correction);
 
   return 1;
 }
@@ -700,10 +696,10 @@ LCL_CanSystemLeap(void)
 /* ================================================== */
 
 void
-LCL_SetSystemLeap(int leap)
+LCL_SetSystemLeap(int leap, int tai_offset)
 {
   if (drv_set_leap) {
-    (drv_set_leap)(leap);
+    (drv_set_leap)(leap, tai_offset);
   }
 }
 
